(in-package "AP5")
#|
Interfaces:

def-process-sentinel (rulename wff)             Macro
  Normally, WFF will be of the form (E <vars> wff'), where wff' is a 2-state
  formula.  For any transaction satisfying this formula, a list of all
  tuples of <vars> for which it is satisfied is associated with the rule.

process-db-wait ({rulename (&optional var) form*}*)               Macro
  the calling process waits until triggering tuples are available for at least
  one of the named rules.  One such rule R is chosen non-deterministically.
  The associated VAR is bound (lexically) to a list of the tuples
  that satisfied R in the earliest transaction having any such tuples
  and not yet consumed by a sentinel.  It is guaranteed that the rule selected
  will be from the earliest transaction for which any of the named rules has 
  as yet unconsumed tuples.  A tuple is represented as a lisp list of the tuple
  elements.  Within the scope of this binding of VAR, the sequence of FORMS is 
  evaluated.  No interesting value is returned from process-db-wait.  Note 
  that, as with other automation rules,
  there is no guarantee that the state of the DB when response FORMs are
  executed bears any particular relationship to its state at the termination
  of the triggering transaction.  It is the programmer's
  responsibility to either guarantee synchronization with locks or waits,
  or to gather enough information in the vars that the response forms do not
  rely on anything else (changeable) about that DB state.



  Typically a process using this capability will be created as:

   (spawn-process "responder"
      #'(lambda ()
          (loop ;;infinite loop
            (process-db-wait
              (siting (tpls) ;; SITING is a rule triggered by a binary relation
                 (atomic (loop for (track loc) in tpls do ...)))
              (redeployment (tpls) 
                ;;REDEPLOYMENT is a rule triggered by a ternary relation
                 (atomic (loop for (ship fleet authorization) in tpls do ...)))))))
  

This implementation allows multiple processes to wait on the SAME sentinel.
When this occurs, there is a race condition.   PROCESS-DB-WAIT ensures
that only one process gets to use any given triggering tuple.  
Would it be better to require a process and all its sentinels to be
declared at once, making it impossible for two processes to wait
on the same sentinel?


|#

(defstruct trigger-buffer
  atomic-id
  (tuples nil :type list))

(defvar *trigger-buffer-lock* nil "lock place for ALL trigger buffers")
 ;; could have a lock per rule, but protected code is always very short

(defmacro with-trigger-buffer-lock (&body b)
  `(lcl:with-process-lock (*trigger-buffer-lock* (current-process)
			    :wait-state "Trigger Buffer Lock")
     ,. b))

(lisp:defun enqueue-process-tuple
    (&rest tpl	&aux (ctpl (copy-list tpl)))
  (with-trigger-buffer-lock
    (let ((tail (last (get-structure-property *current-rule* :trigger-buffer))))
      ;; tail is NIL or a one-element list holding the trigger buffer for
      ;; the most recent atomic with an uncosumed entry for *current-rule*
       (cond
	 ((null tail)
	  ;nothing pending
	  (put-structure-property
	     *current-rule*
	     (list (make-trigger-buffer
		      :atomic-id *atomic-id*
		      :tuples (list ctpl)))
	     :trigger-buffer))
	 ((= (trigger-buffer-atomic-id (first tail)) *atomic-id*)
	  ;; add another tuple for this atomic
	  (push ctpl (trigger-buffer-tuples (first tail))))
	 (t (setf (rest tail)
	       (make-trigger-buffer
		      :atomic-id *atomic-id*
		      :tuples (list ctpl))))))))

(defmacro def-process-sentinel
    (rulename desc &optional documentation)
  `(defautomationgenerator ,rulename ,desc
      enqueue-process-tuple
     ,. (when documentation `(:documentation ,documentation))))


(defmacro process-db-wait (&rest body &aux (which (gensym)))
  ;; WHICH is used as both a tagbody tag and as a local variable.
  ;; If I cared about reading the expansion, I would use two different gensyms.

  `(let (,which)
     (tagbody
       ,which ;; return here in case I lose in a RACE condition
       (process-wait "db-activation"
	 #'(lambda (rules)
	       (setf ,which (has-buffered-tuple rules)))
	 (load-memo (list ,.(loop for (rulename nil . nil) in body
			      collect `(cons (dbo rule ,rulename) ',rulename)))))
       (process-wait "atomic complete"
	  #'(lambda (id) (not (= id *atomic-id*)))
	  (rest ,which))

       (case (first ,which)
       ,.(loop for (rulename tplvar . forms) in body
	   unless (and (consp tplvar)
		       (symbolp (first tplvar))
		       (null (rest tplvar)))
           do (error "syntax error -- bad tuple list variable specification")
	   collect `(,rulename (let ((,(first tplvar)
				      (trigger-buffer-tuples
				       (with-trigger-buffer-lock
					   (let* ((rule (load-memo
							 (dbo rule ,rulename)))
						  (tbuf (get-structure-property
							 rule
							 :trigger-buffer)))
					     (if (and tbuf
						   (= (trigger-buffer-atomic-id
						       (first tbuf))
						      (rest ,which)))
						 ;; nobody else grabbed it!!
						 (pop (get-structure-property
						       rule
						       :trigger-buffer))
	 ;;; a RACE condition, some other process grabbed it first
					       (go ,which)))))))
				  ,.forms)))))))

(lisp:defun has-buffered-tuple (rules) ;; only called from a process-wait
  ;; When there are multiple rules, should this return only a result 
  ;; corresponding to an OLDEST atomic?
  (loop with tmp and earliestid and earliestrulename
   for (rule . rulename) in rules 
   when (setf tmp (get-structure-property rule :trigger-buffer))
   do (when (or (null earliestrulename)
	      (< (the integer (trigger-buffer-atomic-id (first tmp)))
		 (the integer earliestid)))
	  (setf earliestrulename rulename
		earliestid (trigger-buffer-atomic-id (first tmp))))
   finally (when earliestrulename
	     (return (cons rulename (trigger-buffer-atomic-id (first tmp)))))))
